import operator
import pandas as pd
import matplotlib.pyplot as plt
from pandas.core.interchange.dataframe_protocol import DataFrame

from fedex_generator.commons.consts import TOP_K_DEFAULT, DEFAULT_FIGS_IN_ROW
from fedex_generator.commons import utils
from fedex_generator.commons.DatasetRelation import DatasetRelation
from fedex_generator.Operations import Operation
from fedex_generator.Measures.ExceptionalityMeasure import ExceptionalityMeasure
from typing import List, Generator, Tuple

operators = {
    "==": operator.eq,
    ">": operator.gt,
    ">=": operator.ge,
    "<": operator.lt,
    "<=": operator.le,
    "!=": operator.ne,
    "between": lambda x, tup: x.apply(lambda item: tup[0] <= item < tup[1])
}


def do_operation(a, b, op_str):
    return operators[op_str](a, b)


class Filter(Operation.Operation):
    """
    An implementation of the filter operation, fit for the explainability framework.\n
        Provides a .explain() method for explaining the operation, as well as methods used for producing the explanation.
    """

    def __init__(self, source_df: DataFrame, source_scheme: dict, attribute: str = None,
                 operation_str: str = None, value=None, result_df: DataFrame = None):
        """
        :param source_df: The source DataFrame, before the filter operation.
        :param source_scheme: The scheme of the source DataFrame.
        :param attribute: The attribute to filter by.
        :param operation_str: The operation to perform. Only needed if result_df is None.
        :param value: The value to filter by. Only needed if result_df is None.
        :param result_df: The resulting DataFrame after the filter operation.
        :param use_sampling: Whether to use sampling to speed up the generation of explanations. Note that this may
        affect the accuracy and quality of the explanations. Default is True.
        """
        super().__init__(source_scheme)
        self.source_df = source_df.reset_index()
        self.attribute = attribute
        self.source_scheme = source_scheme
        self.cor_deleted_atts = {}
        self.not_presented = {}
        self.corr = self.source_df.corr(numeric_only=True)
        self.type = 'filter'
        self.operation_str = operation_str
        self.value = value

        # If the result_df is not given, we calculate it.
        if result_df is None:
            self.result_df = self.source_df[do_operation(self.source_df[attribute], value, operation_str)]
        else:
            self.result_df = result_df
            self.result_name = utils.get_calling_params_name(result_df)

        self.source_name = utils.get_calling_params_name(source_df)
        self._high_correlated_columns = None
        self._measure = None

    def get_correlated_attributes(self) -> List[str]:
        """
        Get the attributes that are highly correlated with the specified attribute.

        This function calculates the correlation matrix for the numeric columns in the source DataFrame.
        It then identifies and returns the columns that have a correlation coefficient greater than 0.85
        with the specified attribute.

        :return: A list of attributes that are highly correlated with the specified attribute.
        """
        # Avoid repeating the calculation every single time.
        if self._high_correlated_columns is not None:
            return self._high_correlated_columns
        # For performance, we only take the first 10000 rows. We also copy the df because otherwise we
        # are modifying the original df.
        numeric_df = self.source_df.head(10000).copy()

        # For every non-numeric column, we map it to a numeric value by sorting the unique values and assigning
        # them a number.
        for column in numeric_df:
            try:
                if utils.is_numeric(numeric_df[column]):
                    continue

                items = sorted(numeric_df[column].dropna().unique())
                items_map = dict(zip(items, range(len(items))))
                # Changed from numeric_df[column] = numeric_df[column].map(items_map) to avoid SettingWithCopyWarning
                # and future deprecation.
                numeric_df.loc[:, column] = numeric_df[column].map(items_map)
            except Exception as e:
                print(e)

        corr = numeric_df.corr()
        high_correlated_columns = []
        if self.attribute in corr:
            df = corr[self.attribute]

            df = df[df > 0.85].dropna()
            high_correlated_columns = list(df.index)

        self._high_correlated_columns = high_correlated_columns

        return high_correlated_columns

    def iterate_attributes(self) -> Generator[Tuple[str, DatasetRelation], None, None]:
        """
        Iterate over the attributes of the result DataFrame.

        This function yields attributes from the result DataFrame that are not the index,
        the specified attribute, or highly correlated with the specified attribute.
        It also skips attributes marked as 'i' in the source scheme.

        :yield: A tuple containing the attribute name and a DatasetRelation object with the source and result DF.
        """
        high_correlated_columns = self.get_correlated_attributes()

        # Series does not have columns, so if we get here with a series and don't do this we will get an error.
        if not isinstance(self.result_df, pd.Series):
            for attr in self.result_df.columns:
                if isinstance(attr, str) and (attr.lower() == "index" or (self.attribute is not None and attr.lower() == self.attribute.lower()) or \
                                              self.source_scheme.get(attr, None) == 'i' or attr in high_correlated_columns):
                    continue
                yield attr, DatasetRelation(self.source_df, self.result_df, self.source_name)
        else:
            # If we are here, we are dealing with a series.
            if self.attribute is not None and self.attribute.lower() == self.result_df.name.lower():
                return
            yield self.result_df.name, DatasetRelation(self.source_df, self.result_df, self.source_name)

    def get_source_col(self, filter_attr, filter_values, bins) -> pd.Series | None:
        """
        Get the binned column from the source DataFrame for the specified filter attribute and values.

        This function bins the specified attribute in the source DataFrame into the given number of bins and
        returns the binned column that matches the specified filter values.

        :param filter_attr: The attribute to filter by.
        :param filter_values: The values to filter by.
        :param bins: The number of bins to use for binning the attribute.
        :return: The binned column that matches the filter values, or None if the attribute is not in the source DataFrame.
        """
        if filter_attr not in self.source_df:
            return None

        binned_col = pd.cut(self.source_df[filter_attr], bins=bins, labels=False, include_lowest=True,
                            duplicates='drop')

        return binned_col[binned_col.isin(filter_values)]

    def explain(self, schema=None, attributes=None, top_k=TOP_K_DEFAULT,
                figs_in_row: int = DEFAULT_FIGS_IN_ROW, show_scores: bool = False, title: str = None,
                corr_TH: float = 0.7, explainer='fedex', consider='right', cont=None, attr=None, ignore=[],
                use_sampling: bool = True, sample_size=Operation.SAMPLE_SIZE, debug_mode: bool = False,
                draw_figures: bool = True, return_scores: bool = False, measure_only: bool = False)\
            -> (None | Tuple[str, pd.Series, int, int, pd.Series, pd.Series, pd.Series, str, bool] | Tuple):
        """
        Explain for filter operation

        :param schema: dictionary with new columns names, in case {'col_name': 'i'} will be ignored in the explanation
        :param attributes: only this attributes will be included in the explanation calculation
        :param top_k: top k explanations number, default one explanation only.
        :param show_scores: show scores on explanation
        :param figs_in_row: number of explanations figs in one row
        :param title: explanation title
        :param corr_TH: Correlation threshold for deleting correlated attributes
        :param explainer: Which explainer to use. Currently, only 'fedex' is supported for this operation.
        :param consider: Unused but kept for compatibility.
        :param cont: Unused but kept for compatibility.
        :param attr: Unused but kept for compatibility.
        :param ignore: Unused but kept for compatibility.
        :param use_sampling: Whether to use sampling to speed up the generation of explanations.
        :param sample_size: The sample size to use when using sampling. Can be a number or a percentage of the dataframe size. Default is 5000.
        :param debug_mode: Developer option for debugging. Disables multiprocessing when enabled. Can also be used to print additional information. Default is False.
        :param draw_figures: Whether to draw the figures at this stage, or to return the raw explanations, to be drawn later with potentially more added to them. Default is True.
        :param return_scores: Whether to return the scores of the measure or not. Default is False.

        :return: explain figures
        """
        source_df_backup, result_df_backup = None, None

        if use_sampling:
            source_df_backup, result_df_backup = self.source_df, self.result_df
            self.source_df, self.result_df = self.sample(self.source_df, sample_size), self.sample(self.result_df,
                                                                                                   sample_size)

        if attributes is None:
            attributes = []

        if schema is None:
            schema = {}

        # The measure used for filer operations is always the ExceptionalityMeasure.
        measure = ExceptionalityMeasure()
        self._measure = measure
        scores = measure.calc_measure(self, schema, attributes, unsampled_source_df=source_df_backup,
                                      unsampled_res_df=result_df_backup, debug_mode=debug_mode)

        if use_sampling:
            self.source_df, self.result_df = source_df_backup, result_df_backup

        if measure_only:
            return scores

        self.delete_correlated_atts(measure, TH=corr_TH)
        if draw_figures:
            # Get the explanation figures.
            figures = measure.calc_influence(utils.max_key(scores), top_k=top_k, figs_in_row=figs_in_row,
                                             show_scores=show_scores, title=title, debug_mode=debug_mode,
                                             draw_figures=draw_figures)
            if figures:
                self.correlated_notes(figures, top_k)

        else:

            ret_val = measure.calc_influence(utils.max_key(scores), top_k=top_k, figs_in_row=figs_in_row,
                                          show_scores=show_scores, title=title, debug_mode=debug_mode,
                                          draw_figures=draw_figures)
            if return_scores:
                return ret_val, scores
            else:
                return ret_val, None

        if not return_scores:
            return None, None
        else:
            return None, scores

    def draw_figures(self, title: str, scores: pd.Series, K: int, figs_in_row: int, explanations: pd.Series,
                     bins: pd.Series,
                     influence_vals: pd.Series, source_name: str, show_scores: bool,
                     added_text: dict | None = None, added_text_name: str | None = None) -> tuple:
        if self._measure is None:
            raise ValueError(
                "The explain method must be called first before drawing the figures via the draw_figures method.")
        figures, fig = self._measure.draw_figures(
            title=title, scores=scores, K=K, figs_in_row=figs_in_row,
            explanations=explanations, bins=bins, influence_vals=influence_vals,
            source_name=source_name, show_scores=show_scores, added_text=added_text,
            added_text_name=added_text_name
        )
        if figures:
            self.correlated_notes(figures, K, added_text = added_text)

        return figures, fig

    def present_deleted_correlated(self, figs_in_row: int = DEFAULT_FIGS_IN_ROW):
        """
        Present the attributes that were deleted due to high correlation.

        This method calculates the influence of the deleted attributes and generates figures to visualize them.
        It uses the ExceptionalityMeasure to compute the influence values and displays the figures in a grid layout.

        :param figs_in_row: The number of figures to display in one row. Default is the value of DEFAULT_FIGS_IN_ROW.
        """
        if len(self.not_presented) == 0:
            print("No attributes were deleted due to high correlation.")
            return
        measure = ExceptionalityMeasure()
        measure.calc_influence(deleted=self.not_presented, figs_in_row=figs_in_row)

    def correlated_notes(self, figures, top_k, added_text: dict | None = None) -> None:
        """
        Add notes about correlated attributes to the figures.

        This method iterates through the figures and checks for attributes that were deleted due to high correlation.
        If such attributes are found, it adds a note to the figure explaining which attributes were deleted and why.
        It also stores the deleted attributes in the `not_presented` dictionary for later use.

        :param figures: The list of figures to add notes to.
        :param top_k: The number of top attributes to consider.
        :param added_text: The additional text that was already added to the figures, if any. If it exists, the notes will be
        adjusted to be added below the existing text.
        """
        txt = ""
        lentxt = 0
        self.not_presented = {}

        # Go over the figures and check if any of the attributes were deleted due to high correlation.
        for i in range(len(figures)):
            for cor_del in self.cor_deleted_atts.keys():
                # If the attribute was deleted and it is not the last figure, add a note to the figure.
                if figures[i] == cor_del[1] and i < (top_k - 1):
                    lentxt += 1
                    txt += "[" + str(lentxt) + "] " + "The attribute " + cor_del[
                        0] + " is not presented as it correlates with " + cor_del[1] + " (cor: " + str(
                        round(self.corr[cor_del[0]][cor_del[1]], 2)) + ")\n"
                    self.not_presented[cor_del[0]] = self.cor_deleted_atts[cor_del]
        if lentxt > 0:
            txt += "\nIn order to view the not presented attributes, please execute the following: df.present_deleted_correlated()"

            # Add the notes to the figures and show them.
            if added_text is None:
                plt.figtext(0, 0, txt, horizontalalignment='left', verticalalignment='top')
            else:
                # Add "Notes" to the start of the text if there is already text added.
                if not txt.startswith("Notes"):
                    txt = "Notes:\n" + txt
                added_text_num_lines = len(added_text)
                # If there is already text added, we add the notes below the existing text.
                # The extra 0.1 is some extra safety margin, because this is choice of spacing is just a heuristic
                # born of trial and error.
                plt.figtext(0, -0.05 * (added_text_num_lines + 0.1), txt, horizontalalignment='left', verticalalignment='top', fontsize=16)

    def delete_correlated_atts(self, measure, TH=0.7):
        """
        Delete attributes that are highly correlated with each other.

        This method identifies pairs of attributes in the correlation matrix that have a correlation coefficient
        greater than the specified threshold (TH). It then deletes the attribute with the lower measure score
        from the measure's score dictionary.

        :param measure: The measure object containing the score dictionary.
        :param TH: The correlation threshold for deleting attributes. Default is 0.7.
        """

        self.cor_deleted_atts = {}
        corelated_atts = []
        attributes = self.corr.keys()
        numattributes = len(attributes)

        for att in range(numattributes):
            # For each pair of attributes, check if they are highly correlated.
            for att1 in range(att, numattributes):
                cor = self.corr[attributes[att]][attributes[att1]]
                # If they are, and they are not the same attribute, add them to the list of correlated attributes.
                if (cor > TH or cor < -TH) and not att == att1:
                    corelated_atts.append([attributes[att], attributes[att1]])

        # For each pair of correlated attributes, delete the attribute with the lower measure score.
        # Also, store the deleted attribute in the `cor_deleted_atts` dictionary.
        for cor in corelated_atts:
            if (len(set(cor) - set(measure.score_dict))) == 0:
                if measure.score_dict[cor[0]][2] > measure.score_dict[cor[1]][2]:
                    self.cor_deleted_atts[cor[1], cor[0]] = measure.score_dict[cor[1]]
                    del measure.score_dict[cor[1]]
                else:
                    self.cor_deleted_atts[cor[0], cor[1]] = measure.score_dict[cor[0]]
                    del measure.score_dict[cor[0]]



# for performance, we cloned this code from scipy and modified it
import warnings
import math
from math import gcd
import numpy as np
import scipy.special as special
from scipy.stats import distributions
from collections import namedtuple

KstestResult = namedtuple('KstestResult', ('statistic', 'pvalue'))


def _count_paths_outside_method(m, n, g, h):
    """Count the number of paths that pass outside the specified diagonal.

    Parameters
    ----------
    m : integer
        m > 0
    n : integer
        n > 0
    g : integer
        g is greatest common divisor of m and n
    h : integer
        0 <= h <= lcm(m,n)

    Returns
    -------
    p : float
        The number of paths that go low.
        The calculation may overflow - check for a finite answer.

    Raises
    ------
    FloatingPointError: Raised if the intermediate computation goes outside
    the range of a float.

    Notes
    -----
    Count the integer lattice paths from (0, 0) to (m, n), which at some
    point (x, y) along the path, satisfy:
      m*y <= n*x - h*g
    The paths make steps of size +1 in either positive x or positive y
    directions.

    We generally follow Hodges' treatment of Drion/Gnedenko/Korolyuk.
    Hodges, J.L. Jr.,
    "The Significance Probability of the Smirnov Two-Sample Test,"
    Arkiv fiur Matematik, 3, No. 43 (1958), 469-86.

    """
    # Compute #paths which stay lower than x/m-y/n = h/lcm(m,n)
    # B(x, y) = #{paths from (0,0) to (x,y) without
    #             previously crossing the boundary}
    #         = binom(x, y) - #{paths which already reached the boundary}
    # Multiply by the number of path extensions going from (x, y) to (m, n)
    # Sum.

    # Probability is symmetrical in m, n.  Computation below assumes m >= n.
    if m < n:
        m, n = n, m
    mg = m // g
    ng = n // g

    # Not every x needs to be considered.
    # xj holds the list of x values to be checked.
    # Wherever n*x/m + ng*h crosses an integer
    lxj = n + (mg - h) // mg
    xj = [(h + mg * j + ng - 1) // ng for j in range(lxj)]
    # B is an array just holding a few values of B(x,y), the ones needed.
    # B[j] == B(x_j, j)
    if lxj == 0:
        return np.round(special.binom(m + n, n))
    B = np.zeros(lxj)
    B[0] = 1
    # Compute the B(x, y) terms
    # The binomial coefficient is an integer, but special.binom()
    # may return a float. Round it to the nearest integer.
    for j in range(1, lxj):
        Bj = np.round(special.binom(xj[j] + j, j))
        if not np.isfinite(Bj):
            raise FloatingPointError()
        for i in range(j):
            bin = np.round(special.binom(xj[j] - xj[i] + j - i, j - i))
            Bj -= bin * B[i]
        B[j] = Bj
        if not np.isfinite(Bj):
            raise FloatingPointError()
    # Compute the number of path extensions...
    num_paths = 0
    for j in range(lxj):
        bin = np.round(special.binom((m - xj[j]) + (n - j), n - j))
        term = B[j] * bin
        if not np.isfinite(term):
            raise FloatingPointError()
        num_paths += term
    return np.round(num_paths)


def _compute_prob_outside_square(n, h):
    """
    Compute the proportion of paths that pass outside the two diagonal lines.

    Parameters
    ----------
    n : integer
        n > 0
    h : integer
        0 <= h <= n

    Returns
    -------
    p : float
        The proportion of paths that pass outside the lines x-y = +/-h.

    """
    # Compute Pr(D_{n,n} >= h/n)
    # Prob = 2 * ( binom(2n, n-h) - binom(2n, n-2a) + binom(2n, n-3a) - ... )
    # / binom(2n, n)
    # This formulation exhibits subtractive cancellation.
    # Instead divide each term by binom(2n, n), then factor common terms
    # and use a Horner-like algorithm
    # P = 2 * A0 * (1 - A1*(1 - A2*(1 - A3*(1 - A4*(...)))))

    P = 0.0
    k = int(np.floor(n / h))
    while k >= 0:
        p1 = 1.0
        # Each of the Ai terms has numerator and denominator with
        # h simple terms.
        for j in range(h):
            p1 = (n - k * h - j) * p1 / (n + k * h + j + 1)
        P = p1 * (1.0 - P)
        k -= 1
    return 2 * P


def _compute_prob_inside_method(m, n, g, h):
    """
    Count the proportion of paths that stay strictly inside two diagonal lines.

    Parameters
    ----------
    m : integer
        m > 0
    n : integer
        n > 0
    g : integer
        g is greatest common divisor of m and n
    h : integer
        0 <= h <= lcm(m,n)

    Returns
    -------
    p : float
        The proportion of paths that stay inside the two lines.

    Count the integer lattice paths from (0, 0) to (m, n) which satisfy
    |x/m - y/n| < h / lcm(m, n).
    The paths make steps of size +1 in either positive x or positive y
    directions.

    We generally follow Hodges' treatment of Drion/Gnedenko/Korolyuk.
    Hodges, J.L. Jr.,
    "The Significance Probability of the Smirnov Two-Sample Test,"
    Arkiv fiur Matematik, 3, No. 43 (1958), 469-86.

    """
    # Probability is symmetrical in m, n.  Computation below uses m >= n.
    if m < n:
        m, n = n, m
    mg = m // g
    ng = n // g

    # Count the integer lattice paths from (0, 0) to (m, n) which satisfy
    # |nx/g - my/g| < h.
    # Compute matrix A such that:
    #  A(x, 0) = A(0, y) = 1
    #  A(x, y) = A(x, y-1) + A(x-1, y), for x,y>=1, except that
    #  A(x, y) = 0 if |x/m - y/n|>= h
    # Probability is A(m, n)/binom(m+n, n)
    # Optimizations exist for m==n, m==n*p.
    # Only need to preserve a single column of A, and only a
    # sliding window of it.
    # minj keeps track of the slide.
    minj, maxj = 0, min(int(np.ceil(h / mg)), n + 1)
    curlen = maxj - minj
    # Make a vector long enough to hold maximum window needed.
    lenA = min(2 * maxj + 2, n + 1)
    # This is an integer calculation, but the entries are essentially
    # binomial coefficients, hence grow quickly.
    # Scaling after each column is computed avoids dividing by a
    # large binomial coefficient at the end, but is not sufficient to avoid
    # the large dyanamic range which appears during the calculation.
    # Instead we rescale based on the magnitude of the right most term in
    # the column and keep track of an exponent separately and apply
    # it at the end of the calculation.  Similarly when multiplying by
    # the binomial coefficint
    dtype = np.float64
    A = np.zeros(lenA, dtype=dtype)
    # Initialize the first column
    A[minj:maxj] = 1
    expnt = 0
    for i in range(1, m + 1):
        # Generate the next column.
        # First calculate the sliding window
        lastminj, lastlen = minj, curlen
        minj = max(int(np.floor((ng * i - h) / mg)) + 1, 0)
        minj = min(minj, n)
        maxj = min(int(np.ceil((ng * i + h) / mg)), n + 1)
        if maxj <= minj:
            return 0
        # Now fill in the values
        A[0:maxj - minj] = np.cumsum(A[minj - lastminj:maxj - lastminj])
        curlen = maxj - minj
        if lastlen > curlen:
            # Set some carried-over elements to 0
            A[maxj - minj:maxj - minj + (lastlen - curlen)] = 0
        # Rescale if the right most value is over 2**900
        val = A[maxj - minj - 1]
        _, valexpt = math.frexp(val)
        if valexpt > 900:
            # Scaling to bring down to about 2**800 appears
            # sufficient for sizes under 10000.
            valexpt -= 800
            A = np.ldexp(A, -valexpt)
            expnt += valexpt

    val = A[maxj - minj - 1]
    # Now divide by the binomial (m+n)!/m!/n!
    for i in range(1, n + 1):
        val = (val * i) / (m + i)
        _, valexpt = math.frexp(val)
        if valexpt < -128:
            val = np.ldexp(val, -valexpt)
            expnt += valexpt
    # Finally scale if needed.
    return np.ldexp(val, expnt)


def _attempt_exact_2kssamp(n1, n2, g, d, alternative):
    """Attempts to compute the exact 2sample probability.

    n1, n2 are the sample sizes
    g is the gcd(n1, n2)
    d is the computed max difference in ECDFs

    Returns (success, d, probability)
    """
    lcm = (n1 // g) * n2
    h = int(np.round(d * lcm))
    d = h * 1.0 / lcm
    if h == 0:
        return True, d, 1.0
    saw_fp_error, prob = False, np.nan
    try:
        if alternative == 'two-sided':
            if n1 == n2:
                prob = _compute_prob_outside_square(n1, h)
            else:
                prob = 1 - _compute_prob_inside_method(n1, n2, g, h)
        else:
            if n1 == n2:
                # prob = binom(2n, n-h) / binom(2n, n)
                # Evaluating in that form incurs roundoff errors
                # from special.binom. Instead calculate directly
                jrange = np.arange(h)
                prob = np.prod((n1 - jrange) / (n1 + jrange + 1.0))
            else:
                num_paths = _count_paths_outside_method(n1, n2, g, h)
                bin = special.binom(n1 + n2, n1)
                if not np.isfinite(bin) or not np.isfinite(num_paths) \
                        or num_paths > bin:
                    saw_fp_error = True
                else:
                    prob = num_paths / bin

    except FloatingPointError:
        saw_fp_error = True

    if saw_fp_error:
        return False, d, np.nan
    if not (0 <= prob <= 1):
        return False, d, prob
    return True, d, prob


def get_unique_numbers(numbers):
    object_numbers = np.array(numbers, dtype=object)
    shifted_numbers = np.append(object_numbers[1:], np.nan)

    return numbers[object_numbers != shifted_numbers]


def ks_2samp(data1, data2, alternative='two-sided', mode='asymp'):
    """
    Performs the two-sample Kolmogorov-Smirnov test for goodness of fit.

    This test compares the underlying continuous distributions F(x) and G(x)
    of two independent samples.  See Notes for a description
    of the available null and alternative hypotheses.

    Parameters
    ----------
    data1, data2 : array_like, 1-Dimensional
        Two arrays of sample observations assumed to be drawn from a continuous
        distribution, sample sizes can be different.
    alternative : {'two-sided', 'less', 'greater'}, optional
        Defines the null and alternative hypotheses. Default is 'two-sided'.
        Please see explanations in the Notes below.
    mode : {'auto', 'exact', 'asymp'}, optional
        Defines the method used for calculating the p-value.
        The following options are available (default is 'auto'):

          * 'auto' : use 'exact' for small size arrays, 'asymp' for large
          * 'exact' : use exact distribution of test statistic
          * 'asymp' : use asymptotic distribution of test statistic

    Returns
    -------
    statistic : float
        KS statistic.
    pvalue : float
        One-tailed or two-tailed p-value.

    See Also
    --------
    kstest, ks_1samp, epps_singleton_2samp, anderson_ksamp

    Notes
    -----
    There are three options for the null and corresponding alternative
    hypothesis that can be selected using the `alternative` parameter.

    - `two-sided`: The null hypothesis is that the two distributions are
      identical, F(x)=G(x) for all x; the alternative is that they are not
      identical.

    - `less`: The null hypothesis is that F(x) >= G(x) for all x; the
      alternative is that F(x) < G(x) for at least one x.

    - `greater`: The null hypothesis is that F(x) <= G(x) for all x; the
      alternative is that F(x) > G(x) for at least one x.

    Note that the alternative hypotheses describe the *CDFs* of the
    underlying distributions, not the observed values. For example,
    suppose x1 ~ F and x2 ~ G. If F(x) > G(x) for all x, the values in
    x1 tend to be less than those in x2.


    If the KS statistic is small or the p-value is high, then we cannot
    reject the null hypothesis in favor of the alternative.

    If the mode is 'auto', the computation is exact if the sample sizes are
    less than 10000.  For larger sizes, the computation uses the
    Kolmogorov-Smirnov distributions to compute an approximate value.

    The 'two-sided' 'exact' computation computes the complementary probability
    and then subtracts from 1.  As such, the minimum probability it can return
    is about 1e-16.  While the algorithm itself is exact, numerical
    errors may accumulate for large sample sizes.   It is most suited to
    situations in which one of the sample sizes is only a few thousand.

    We generally follow Hodges' treatment of Drion/Gnedenko/Korolyuk [1]_.

    References
    ----------
    .. [1] Hodges, J.L. Jr.,  "The Significance Probability of the Smirnov
           Two-Sample Test," Arkiv fiur Matematik, 3, No. 43 (1958), 469-86.

    Examples
    --------
    >>> from scipy import stats
    >>> rng = np.random.default_rng()

    >>> n1 = 200  # size of first sample
    >>> n2 = 300  # size of second sample

    For a different distribution, we can reject the null hypothesis since the
    pvalue is below 1%:

    >>> rvs1 = stats.norm.rvs(size=n1, loc=0., scale=1, random_state=rng)
    >>> rvs2 = stats.norm.rvs(size=n2, loc=0.5, scale=1.5, random_state=rng)
    >>> stats.ks_2samp(rvs1, rvs2)
     KstestResult(statistic=0.24833333333333332, pvalue=5.846586728086578e-07)

    For a slightly different distribution, we cannot reject the null hypothesis
    at a 10% or lower alpha since the p-value at 0.144 is higher than 10%

    >>> rvs3 = stats.norm.rvs(size=n2, loc=0.01, scale=1.0, random_state=rng)
    >>> stats.ks_2samp(rvs1, rvs3)
    KstestResult(statistic=0.07833333333333334, pvalue=0.4379658456442945)

    For an identical distribution, we cannot reject the null hypothesis since
    the p-value is high, 41%:

    >>> rvs4 = stats.norm.rvs(size=n2, loc=0.0, scale=1.0, random_state=rng)
    >>> stats.ks_2samp(rvs1, rvs4)
    KstestResult(statistic=0.12166666666666667, pvalue=0.05401863039081145)

    """
    if mode not in ['auto', 'exact', 'asymp']:
        raise ValueError(f'Invalid value for mode: {mode}')
    alternative = {'t': 'two-sided', 'g': 'greater', 'l': 'less'}.get(
        alternative.lower()[0], alternative)
    if alternative not in ['two-sided', 'less', 'greater']:
        raise ValueError(f'Invalid value for alternative: {alternative}')
    MAX_AUTO_N = 10000  # 'auto' will attempt to be exact if n1,n2 <= MAX_AUTO_N
    # np.random.shuffle(data1)
    # np.random.shuffle(data2)
    # data1 = data1[:5000]
    # data2 = data2[:5000]
    if np.ma.is_masked(data1):
        data1 = data1.compressed()
    if np.ma.is_masked(data2):
        data2 = data2.compressed()
    data1 = np.sort(data1)
    data2 = np.sort(data2)
    n1 = data1.shape[0]
    n2 = data2.shape[0]
    if min(n1, n2) == 0:
        raise ValueError('Data passed to ks_2samp must not be empty')

    data_all = np.unique(np.concatenate([get_unique_numbers(data1), get_unique_numbers(data2)]))
    # using searchsorted solves equal data problem
    cdf1 = np.searchsorted(data1, data_all, side='right') / n1
    cdf2 = np.searchsorted(data2, data_all, side='right') / n2
    cddiffs = cdf1 - cdf2
    # Ensure sign of minS is not negative.
    minS = np.clip(-np.min(cddiffs), 0, 1)
    maxS = np.max(cddiffs)
    alt2Dvalue = {'less': minS, 'greater': maxS, 'two-sided': max(minS, maxS)}
    d = alt2Dvalue[alternative]
    g = gcd(n1, n2)
    n1g = n1 // g
    n2g = n2 // g
    prob = -np.inf
    original_mode = mode
    if mode == 'auto':
        mode = 'exact' if max(n1, n2) <= MAX_AUTO_N else 'asymp'
    elif mode == 'exact':
        # If lcm(n1, n2) is too big, switch from exact to asymp
        if n1g >= np.iinfo(np.int32).max / n2g:
            mode = 'asymp'
            warnings.warn(
                f"Exact ks_2samp calculation not possible with samples sizes "
                f"{n1} and {n2}. Switching to 'asymp'.", RuntimeWarning)

    if mode == 'exact':
        success, d, prob = _attempt_exact_2kssamp(n1, n2, g, d, alternative)
        if not success:
            mode = 'asymp'
            if original_mode == 'exact':
                warnings.warn(f"ks_2samp: Exact calculation unsuccessful. "
                              f"Switching to mode={mode}.", RuntimeWarning)

    if mode == 'asymp':
        # The product n1*n2 is large.  Use Smirnov's asymptoptic formula.
        # Ensure float to avoid overflow in multiplication
        # sorted because the one-sided formula is not symmetric in n1, n2
        m, n = sorted([float(n1), float(n2)], reverse=True)
        en = m * n / (m + n)
        if alternative == 'two-sided':
            prob = distributions.kstwo.sf(d, np.round(en))
        else:
            z = np.sqrt(en) * d
            # Use Hodges' suggested approximation Eqn 5.3
            # Requires m to be the larger of (n1, n2)
            expt = -2 * z ** 2 - 2 * z * (m + 2 * n) / np.sqrt(m * n * (m + n)) / 3.0
            prob = np.exp(expt)

    prob = np.clip(prob, 0, 1)
    return KstestResult(d, prob)
